##############################################################################
#
# Zope Public License (ZPL) Version 0.9.4
# ---------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
# 
# 1. Redistributions in source code must retain the above
#    copyright notice, this list of conditions, and the following
#    disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above
#    copyright notice, this list of conditions, and the following
#    disclaimer in the documentation and/or other materials
#    provided with the distribution.
# 
# 3. Any use, including use of the Zope software to operate a
#    website, must either comply with the terms described below
#    under "Attribution" or alternatively secure a separate
#    license from Digital Creations.
# 
# 4. All advertising materials, documentation, or technical papers
#    mentioning features derived from or use of this software must
#    display the following acknowledgement:
# 
#      "This product includes software developed by Digital
#      Creations for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
# 5. Names associated with Zope or Digital Creations must not be
#    used to endorse or promote products derived from this
#    software without prior written permission from Digital
#    Creations.
# 
# 6. Redistributions of any form whatsoever must retain the
#    following acknowledgment:
# 
#      "This product includes software developed by Digital
#      Creations for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
# 7. Modifications are encouraged but must be packaged separately
#    as patches to official Zope releases.  Distributions that do
#    not clearly separate the patches from the original work must
#    be clearly labeled as unofficial distributions.
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND
#   ANY EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
#   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
#   FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT
#   SHALL DIGITAL CREATIONS OR ITS CONTRIBUTORS BE LIABLE FOR ANY
#   DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
#   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
#   DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#   LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
#   IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
#   THE POSSIBILITY OF SUCH DAMAGE.
# 
# Attribution
# 
#   Individuals or organizations using this software as a web site
#   must provide attribution by placing the accompanying "button"
#   and a link to the accompanying "credits page" on the website's
#   main entry point.  In cases where this placement of
#   attribution is not feasible, a separate arrangment must be
#   concluded with Digital Creations.  Those using the software
#   for purposes other than web sites must provide a corresponding
#   attribution in locations that include a copyright using a
#   manner best suited to the application environment.
# 
# This software consists of contributions made by Digital
# Creations and many individuals on behalf of Digital Creations.
# Specific attributions are listed in the accompanying credits
# file.
# 
##############################################################################

"""User Db product

Authenticates off of an /etc/passwd like file.  This file must be in
the etcUsers directory in the top level Zope directory.  etcUserFolder 
only uses the first two columns of a password file, so files generated 
by htpasswd should work also.  The format must be:

uid:crypted_password

"""

import Globals, App.Undo, socket, regex, os

from Globals import HTMLFile, MessageDialog, INSTANCE_HOME
from string import join,strip,split,lower,upper,find
from App.Management import Navigation, Tabs
from Acquisition import Implicit
from OFS.SimpleItem import Item
from OFS.Document import Document
from base64 import decodestring, encodestring
from urllib import quote, unquote
from ImageFile import ImageFile
from AccessControl.Role import RoleManager
from AccessControl.User import BasicUser, BasicUserFolder
from crypt import crypt

path_split=os.path.split
path_join=os.path.join
exists=os.path.exists

class User(BasicUser):
    
    icon='misc_/etcUserFolder/User_icon'

    def __init__(self, name, password, roles, domains):
	self.name   =name
	self.__     =password
        self.roles  =filter(None, map(strip, split(roles, ',')))
        self.domains=filter(None, map(strip, split(domains, ',')))

    def getUserName(self):
        return self.name

    def _getPassword(self):
        return self.__

    def getRoles(self):
        return self.roles
    
    def getDomains(self):
        return self.domains

    def authenticate(self, password, pwfile, request):
        domains=self.getDomains()
	users = []
	pf = open(getPath('etcUsers', pwfile), 'r')
	while 1:
	    n = pf.readline()
	    if not n:
		break
	    users.append(n)
	
	for people in users:
	    vals = split(people, ':')
	    if vals[0] == self.name:
		if vals[1][-1] == '\n':
		    pwd = vals[1][:-1]
		else:
		    pwd = vals[1]
		salt = vals[1][:2]
		secret = crypt(password, salt)
		break
        if domains:
            return (secret==pwd) and domainSpecMatch(domains, request)
        return secret==pwd


class etcUserFolder(BasicUserFolder):
    """ """

    meta_type='etcUserFolder'
    id       ='acl_users'
    title    ='etcUserFolder'
    icon     ='misc_/etcUserFolder/etc_icon'

    isPrincipiaFolderish=1
    isAUserFolder=1

    manage_options=(
    {'label':'Contents',   'action':'manage_main'},
    {'label':'Properties', 'action':'manage_properties'},
    {'label':'Security',   'action':'manage_access'},
    {'label':'Undo',       'action':'manage_UndoForm'},
    )

    __ac_permissions__=(
    ('View management screens',
     ['manage','manage_menu','manage_main','manage_copyright', 'manage_tabs',
      'manage_properties', 'manage_UndoForm']),
    ('Undo changes',       ['manage_undo_transactions']),
    ('Change permissions', ['manage_access']),
    ('Manage users',       ['manage_users']),
    ('Change etc User Folders',     ['manage_edit']),
    )

    manage_properties=HTMLFile('properties', globals())

    manage_main=HTMLFile('mainUser', globals())

    def __init__(self, pwfile='etcUsers', default_role='etcUser',
		 cookie_mode=0, REQUEST=None):
        self.pwfile=pwfile
        self.cookie_mode=cookie_mode
	self.default_role=default_role

        
        self.docLogin =Document(_docLogin, __name__='docLogin')
        self.docLogout=Document(_docLogout, __name__='docLogout')
        self.docLogin.__roles__=None
        self.docLogout.__roles__=None

	# register default_role with the container folder
	REQUEST['PARENTS'][0]._addRole(default_role, REQUEST)
	

    def listusers(self):
	users = []
	pf = open(getPath('etcUsers', self.pwfile), 'r')
	while 1:
	    n = pf.readline()
	    if not n:
		break
	    users.append(n)
	return users

    def getUserNames(self):
        """Returns a list of user names or [] if no users exist"""
        data=[]
        try:    items=self.listusers()
        except: return data
        for people in items:
	    vals = split(people, ':')
	    name = vals[0]
            data.append(name)
        data.sort()
        return data

    def getUsers(self):
        """Return a list of user objects or [] if no users exist"""
        data=[]
        try:    items=self.listusers()
        except: return data
        for people in items:
	    vals = split(people, ':')
	    #make up a role and no domain
            user=User(vals[0], vals[1], self.default_role, '')
            data.append(user)
        return data

    def getUser(self, name):
        """Return the named user object or None if no such user exists"""
        try:    items=self.listusers()
        except: return None
        if not items:
            return None
	for people in items:
	    vals = split(people, ':')
	    if vals[0] == name:
		return User(vals[0], vals[1], self.default_role, '')
	return None


    def std_validate(self,request,auth='',roles=None):
        parent=request['PARENTS'][0]

        # If no authorization, only a user with a
        # domain spec and no passwd or nobody can
        # match
        if not auth:
            for ob in self.getUsers():
                domains=ob.getDomains()
                if domains:
                    if ob.authenticate('', request):
                        if ob.allowed(parent, roles):
                            ob=ob.__of__(self)
                            return ob
            nobody=self._nobody
            if self._isTop() and nobody.allowed(parent, roles):
                ob=nobody.__of__(self)
                return ob
            return None

        # Only do basic authentication
        if lower(auth[:6])!='basic ':
            return None
        name,password=tuple(split(decodestring(split(auth)[-1]), ':'))

        # Check for superuser
        super=self._super
        if self._isTop() and (name==super.getUserName()) and \
        super.authenticate(password, request):
            return super

        # Try to get user
        user=self.getUser(name)
        if user is None:
            return None

        # Try to authenticate user
        if not user.authenticate(password, self.pwfile, request):
            return None

        # We need the user to be able to acquire!
        user=user.__of__(self)

        # Try to authorize user
        if user.allowed(parent, roles):
            return user
        return None


    def cookie_validate(self,request,auth='',roles=None):
	parent=request['PARENTS'][0]
        req_has=request.has_key
        resp=request['RESPONSE']
        
        if req_has('__ac'):
            c=request['__ac']
            c=unquote(c)
            try: c=decodestring(c)
            except:
                resp.expireCookie('__ac', path='/')
                raise 'LoginRequired', self.docLogin(self, request)
            
            name,password=tuple(split(c, ':'))
            
            # Find user
            user=self.getUser(name)
            if user is not None:
                if user.authenticate(password, self.pwfile, request):
                    user=user.__of__(self)
                    if user.allowed(parent, roles):
                        return user
            resp.expireCookie('__ac', path='/')
            raise 'LoginRequired', self.docLogin(self, request)

        if req_has('__ac_name') and req_has('__ac_password'):
            name=request['__ac_name']
            password=request['__ac_password']

            # Find user
            user=self.getUser(name)
            if user is not None:
                if user.authenticate(password, self.pwfile, request):
                    user=user.__of__(self)
                    if user.allowed(parent, roles):
                        token='%s:%s' % (name, password)
                        token=encodestring(token)
                        token=quote(token)
                        resp.setCookie('__ac', token, path='/')
                        request['__ac']=token
                        try:
                            del request['__ac_name']
                            del request['__ac_password']
                        except:
                            pass
                        return user
            raise 'LoginRequired', self.docLogin(self, request)
        
        if auth: return None

        # Try domain matching
        for ob in self.getUsers():
            domains=ob.getDomains()
            if ob.domains:
                if ob.authenticate('', self.pwfile, request):
                    if ob.allowed(parent, roles):
                        ob=ob.__of__(self)
                        return ob

        # Allow anonymous access to things under etcUserFolders,
        # if the object is public.
        nobody=self._nobody
        if nobody.allowed(parent, roles):
            ob=nobody.__of__(self)
            return ob

        raise 'LoginRequired', self.docLogin(self, request)


    def validate(self,request,auth='',roles=None):
        if self.cookie_mode:
            return self.cookie_validate(request, auth, roles)
        return self.std_validate(request, auth, roles)
    

    def __len__(self):
        return len(self.listusers())


    def manage_edit(self, pwfile, default_role, cookie_mode, REQUEST=None):
	"""Change properties"""

        self.cookie_mode=cookie_mode
	self.pwfile = pwfile
	self.default_role = default_role


        if REQUEST:
            return MessageDialog(
                   title  ='etcUserFolder Changed',
                   message='UserDb properties have been updated',
                   action =REQUEST['URL1']+'/manage_properties',
                   target ='manage_main')

    logout__roles__=None

    def logout(self, REQUEST):
        """Logout"""
        REQUEST['RESPONSE'].expireCookie('__ac', path='/')
        return self.docLogout(self, REQUEST)
    

manage_addetcUserFolderForm=HTMLFile('addetcUserFolder', globals())

def manage_addetcUserFolder(self, pwfile='etcUsers', 
			    default_role='etcUser',
			    cookie_mode=0, REQUEST=None):
    """ """
    if hasattr(self.aq_base, 'acl_users'):
        return MessageDialog(
               title  ='Item Exists',
               message='This object already contains a User Folder',
               action ='%s/manage_main' % REQUEST['PARENT_URL'])
    ob=etcUserFolder(pwfile, default_role, cookie_mode, REQUEST)
    self._setObject('acl_users', ob)
    self.__allow_groups__=self.acl_users
    if REQUEST: return self.manage_main(self,REQUEST,update_menu=1)

def _getPath(home, prefix, name, suffixes):
    d=path_join(home, prefix)
    if d==prefix: raise ValueError, (
        'The prefix, %s, should be a relative path' % prefix)
    d=path_join(d,name)
    if d==name: raise ValueError, ( # Paranoia
        'The file name, %s, should be a simple file name' % name)
    for s in suffixes:
        if s: s="%s.%s" % (d, s)
        else: s=d
        if exists(s): return s

def getPath(prefix, name, checkProduct=1, suffixes=('',)):
    """Find a file in one of several relative locations

    Arguments:

      prefix -- The location, relative to some home, to look for the
                file

      name -- The name of the file.  This must not be a path.

      checkProduct -- a flag indicating whether product directories
        should be used as additional hope ares to be searched. This
        defaults to a true value.

        If this is true and the name contains a dot, then the
        text before the dot is treated as a product name and
        the product package directory is used as anothe rhome.

      suffixes -- a sequences of file suffixes to check.
        By default, the name is used without a suffix.

    The search takes on multiple homes which are INSTANCE_HOME,
    the directory containing the directory containing SOFTWARE_HOME, and
    possibly product areas.     
    """
    d,n = path_split(name)
    if d: raise ValueError, (
        'The file name, %s, should be a simple file name' % name)

    sw=path_split(path_split(SOFTWARE_HOME)[0])[0]
    for home in (INSTANCE_HOME, sw):
        if checkProduct:
            l=find(name, '.')
            if l > 0:
                p=name[:l]
                n=name[l+1:]
                r=_getPath(home, "Products/%s/%s/" % (p,prefix),
                           n, suffixes)
                if r is not None: return r
        r=_getPath(home, prefix, name, suffixes)
        if r is not None: return r



addr_match=regex.compile('[0-9\.\*]*').match
host_match=regex.compile('[A-Za-z0-9\.\*]*').match

def domainSpecValidate(spec):
    for ob in spec:
        sz=len(ob)
        if not ((addr_match(ob) == sz) or (host_match(ob) == sz)):
            return 0
    return 1


def domainSpecMatch(spec, request):

    if request.has_key('REMOTE_HOST'):
        host=request['REMOTE_HOST']
    else: host=''

    if request.has_key('REMOTE_ADDR'):
        addr=request['REMOTE_ADDR']
    else: addr=''

    if not host and not addr:
        return 0

    if not host:
        host=socket.gethostbyaddr(addr)[0]
    if not addr:
        addr=socket.gethostbyname(host)

    _host=split(host, '.')
    _addr=split(addr, '.')
    _hlen=len(_host)
    _alen=len(_addr)
    
    for ob in spec:
        sz=len(ob)
        _ob=split(ob, '.')
        _sz=len(_ob)

        if addr_match(ob)==sz:
            if _sz != _alen:
                continue
            fail=0
            for i in range(_sz):
                a=_addr[i]
                o=_ob[i]
                if (o != a) and (o != '*'):
                    fail=1
                    break
            if fail:
                continue
            return 1

        if host_match(ob)==sz:
            if _hlen < _sz:
                continue
            elif _hlen > _sz:
                _item=_host[-_sz:]
            else:
                _item=_host
            fail=0
            for i in range(_sz):
                h=_item[i]
                o=_ob[i]
                if (o != h) and (o != '*'):
                    fail=1
                    break
            if fail:
                continue
            return 1
    return 0

_docLogin="""<HTML>
<HEAD>
<TITLE>Login</TITLE>
</HEAD>
<BODY BGCOLOR="#FFFFFF" LINK="#000099" VLINK="#555555">
<P>
<FORM ACTION="<!--#var SCRIPT_NAME--><!--#var PATH_INFO--><!--#if QUERY_STRING-->?<!--#var QUERY_STRING--><!--#/if-->" METHOD="POST">
<TABLE>
<TR>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <STRONG>Name</STRONG>
  </TD>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <INPUT TYPE="TEXT" NAME="__ac_name" SIZE="20">
  </TD>
</TR>

<TR>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <STRONG>Password</STRONG>
  </TD>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <INPUT TYPE="PASSWORD" NAME="__ac_password" SIZE="20">
  </TD>
</TR>
<TR>
  <TD ALIGN="LEFT" VALIGN="TOP">
  </TD>
  <TD ALIGN="LEFT" VALIGN="TOP">
  <INPUT TYPE="SUBMIT" NAME="submit" VALUE=" Ok ">
  </TD>
</TR>
</TABLE>
</FORM>
</BODY>
</HTML>"""

_docLogout="""<HTML>
<HEAD>
<TITLE>Logout</TITLE>
</HEAD>
<BODY BGCOLOR="#FFFFFF" LINK="#000099" VLINK="#555555">
<P>
<CENTER>
You have been logged out of the system.
</CENTER>
</BODY>
</HTML>"""










